@@ -25,6 +25,18 @@ swf_tag_shape_detail_handler(void) {
     return &shape_detail_handler;
 }
 
+swf_tag_detail_handler_t *
+swf_tag_shape_cid_handler(void) {
+    shape_detail_handler.create   = NULL;
+    shape_detail_handler.input    = NULL;
+    shape_detail_handler.get_cid     = swf_tag_shape_get_cid_detail;
+    shape_detail_handler.replace_cid = swf_tag_shape_replace_cid_detail;
+    shape_detail_handler.output   = NULL;
+    shape_detail_handler.print    = NULL;
+    shape_detail_handler.destroy  = NULL;
+    return &shape_detail_handler;
+}
+
 void *
 swf_tag_shape_create_detail(void) {
     swf_tag_shape_detail_t *swf_tag_shape;
@@ -65,9 +77,9 @@ swf_tag_shape_input_detail(swf_tag_t *tag, struct swf_object_ *swf) {
     }
 
     // DefineMorphShape, DefineMorphShape2
-    swf_tag_shape->is_morph = (tag->tag == 46) || (tag->tag == 84);
+    swf_tag_shape->is_morph = (tag->code == 46) || (tag->code == 84);
     // DefineShape4, DefineMorphShape2
-    swf_tag_shape->has_strokes = (tag->tag == 83) || (tag->tag == 84);
+    swf_tag_shape->has_strokes = (tag->code == 83) || (tag->code == 84);
 
     if (swf_tag_shape->is_morph) {
         ret = swf_rect_parse(bs, &(swf_tag_shape->rect_morph));
@@ -86,28 +98,34 @@ swf_tag_shape_input_detail(swf_tag_t *tag, struct swf_object_ *swf) {
         }
         if (swf_tag_shape->is_morph) {
             ret = swf_rect_parse(bs, &(swf_tag_shape->stroke_rect_morph));
-	    if (ret) {
-	        fprintf(stderr, "ERROR: swf_tag_shape_input_detail: swf_tag_shape->stroke_rect_morph parse failed. shape_id=%d\n", swf_tag_shape->shape_id);
-	        bitstream_close(bs);
-	        return ret;
-	    }
+            if (ret) {
+                fprintf(stderr, "ERROR: swf_tag_shape_input_detail: swf_tag_shape->stroke_rect_morph parse failed. shape_id=%d\n", swf_tag_shape->shape_id);
+                bitstream_close(bs);
+                return ret;
+            }
         }
         swf_tag_shape->define_shape_reserved = bitstream_getbits(bs, 6);
         swf_tag_shape->define_shape_non_scaling_strokes = bitstream_getbits(bs, 1);
         swf_tag_shape->define_shape_scaling_strokes = bitstream_getbits(bs, 1);
     }
     if (swf_tag_shape->is_morph) {
+        bitstream_align(bs);
         swf_tag_shape->offset_morph = bitstream_getbytesLE(bs, 4);
-        swf_morph_shape_with_style_parse(bs, &swf_tag_shape->morph_shape_with_style, tag);
+        ret = swf_morph_shape_with_style_parse(bs, &swf_tag_shape->morph_shape_with_style, tag);
+        if (ret) {
+            fprintf(stderr, "ERROR: swf_tag_shape_input_detail: swf_shape_with_style_parse swf_tag_shape->morph_shape_with_style failed. shape_id=%d\n", swf_tag_shape->shape_id);
+	        bitstream_close(bs);
+            return ret;
+        }
     } else {
         ret = swf_shape_with_style_parse(bs, &swf_tag_shape->shape_with_style, tag);
         if (ret) {
             fprintf(stderr, "ERROR: swf_tag_shape_input_detail: swf_shape_with_style_parse swf_tag_shape->shape_with_style failed. shape_id=%d\n", swf_tag_shape->shape_id);
 	        bitstream_close(bs);
-                return ret;
+            return ret;
         }
     }
-        
+    
     bitstream_close(bs);
     return 0;
 }
@@ -137,70 +155,131 @@ int swf_tag_shape_replace_cid_detail(swf_tag_t *tag, int id) {
     return 0; // always 0
 }
 
-int swf_tag_shape_bitmap_get_refcid(swf_tag_t *tag) {
+static swf_shape_record_t *
+_swf_tag_shape_search_new_style_in_shape_records(swf_shape_record_t *shape_records) {
+    for ( ; shape_records ; shape_records = shape_records->next) {
+        if ((shape_records->first_6bits) && 
+            ((shape_records->first_6bits & 0x20) == 0)) {
+            if (shape_records->shape.shape_setup.shape_has_new_styles) {
+                break;
+            }
+        }
+    }
+    return shape_records;
+}
+
+int *
+swf_tag_shape_bitmap_get_refcid_list(swf_tag_t *tag, int *cid_list_num) {
     swf_tag_shape_detail_t *swf_tag_shape;
     int i, ret;
+    int *cid_list, cid_list_alloc;
+    swf_styles_t *styles = NULL;
+    swf_shape_record_t *shape_records = NULL;
+
     if (tag == NULL) {
         fprintf(stderr, "swf_tag_shape_bitmap_get_refcid: tag == NULL\n");
-        return -1;
+        return NULL;
     }
-    if (! isShapeTag(tag->tag)) {
+    if (! isShapeTag(tag->code)) {
         fprintf(stderr, "swf_tag_shape_bitmap_get_refcid: ! isShapeTag(%d)\n",
-                tag->tag);
-        return -1;
+                tag->code);
+        return NULL;
     }
     if (tag->detail == NULL) {
         tag->detail = swf_tag_shape_create_detail();
         swf_tag_shape = (swf_tag_shape_detail_t *) tag->detail;
-        swf_tag_shape->_parse_condition = SWF_TAG_SHAPE_PARSE_CONDITION_BITMAP;
+	//        swf_tag_shape->_parse_condition = SWF_TAG_SHAPE_PARSE_CONDITION_BITMAP;
         ret = swf_tag_shape_input_detail(tag, NULL);
         if (ret) {
             swf_tag_shape_destroy_detail(tag);
-            return -1; // no shape bitmap
+            return NULL; // no shape bitmap
         }
     } else {
         swf_tag_shape = (swf_tag_shape_detail_t *) tag->detail;
     }
     //
-    for (i = 0 ; i < swf_tag_shape->shape_with_style.styles.fill_styles.count ; i++) {
-        swf_fill_style_t *fill_style;
-        fill_style = &(swf_tag_shape->shape_with_style.styles.fill_styles.fill_style[i]);
-        if (fill_style == NULL) {
-            fprintf(stderr, "swf_tag_shape_bitmap_get_refcid: fill_style == NULL i=%d\n", i);
-            return -1; // Illegal!!!
-        }
-        switch (fill_style->type) {
-          case 0x40: // tilled  bitmap fill with smoothed edges
-          case 0x41: // clipped bitmap fill with smoothed edges
-          case 0x42: // tilled  bitmap fill with hard edges
-          case 0x43: // clipped bitmap fill with hard edges
-            if (fill_style->bitmap.bitmap_ref != 0xffff) {
-                return fill_style->bitmap.bitmap_ref;
+    *cid_list_num = 0;
+    cid_list_alloc = 10;
+    cid_list = malloc(sizeof(int) * cid_list_alloc);
+
+    if (tag->code == 46) { // DefineMorphShape
+        styles = &(swf_tag_shape->morph_shape_with_style.styles);
+        shape_records = &(swf_tag_shape->morph_shape_with_style.shape_records);
+    } else {
+        styles = &(swf_tag_shape->shape_with_style.styles);
+        shape_records = &(swf_tag_shape->shape_with_style.shape_records);
+    }
+    while (1) {
+        for (i = 0 ; i < styles->fill_styles.count ; i++) {
+            swf_fill_style_t *fill_style;
+            fill_style = &(styles->fill_styles.fill_style[i]);
+            if (fill_style == NULL) {
+                fprintf(stderr, "swf_tag_shape_bitmap_get_refcid: fill_style == NULL i=%d\n", i);
+                free(cid_list);
+                return NULL; // Illegal!!!
             }
-            break;
-          default:
-            break;
+            switch (fill_style->type) {
+                void *tmp;
+              case 0x40: // tilled  bitmap fill with smoothed edges
+              case 0x41: // clipped bitmap fill with smoothed edges
+              case 0x42: // tilled  bitmap fill with hard edges
+              case 0x43: // clipped bitmap fill with hard edges
+                if (fill_style->bitmap.bitmap_ref != 0xffff) {
+                    if (cid_list_alloc <= *cid_list_num) {
+                        cid_list_alloc *= 2;
+                        tmp = realloc(cid_list, cid_list_alloc);
+                        if (tmp == NULL) {
+                            fprintf(stderr, "swf_tag_shape_bitmap_get_refcid_list: Can't realloc memory (%p, %d)\n", cid_list, cid_list_alloc);
+                            free(cid_list);
+                            return NULL;
+                        }
+                        cid_list = tmp;
+                    }
+                    cid_list[*cid_list_num] = fill_style->bitmap.bitmap_ref;
+                    *cid_list_num  = (*cid_list_num) + 1;
+                }
+                break;
+              default:
+                break;
+            }
+        }
+        // new style を探す
+        shape_records = _swf_tag_shape_search_new_style_in_shape_records(shape_records);
+        
+        if (shape_records) {
+            styles = &(shape_records->shape.shape_setup.styles);
+            shape_records = shape_records->next; // next
+        } else {
+            break; // finish
         }
     }
-    return -1; // not found
+    
+    if (*cid_list_num == 0) {
+        free(cid_list);
+        return NULL;
+    }
+    return cid_list; // not found
 }
 
-int swf_tag_shape_bitmap_replace_refcid(swf_tag_t *tag, int cid) {
+int swf_tag_shape_bitmap_replace_refcid_list(swf_tag_t *tag, int from_cid, int to_cid) {
     swf_tag_shape_detail_t *swf_tag_shape;
     int i, ret;
+    swf_styles_t *styles = NULL;
+    swf_shape_record_t *shape_records = NULL;
+    int morph_shape_check = 0;
     if (tag == NULL) {
         fprintf(stderr, "swf_tag_shape_bitmap_replace_refcid: tag == NULL\n");
         return 1;
     }
-    if (! isShapeTag(tag->tag)) {
+    if (! isShapeTag(tag->code)) {
         fprintf(stderr, "swf_tag_shape_bitmap_replace_refcid: ! isShapeTag(%d)\n",
-                tag->tag);
+                tag->code);
         return 1;
     }
     if (tag->detail == NULL) {
         tag->detail = swf_tag_shape_create_detail();
         swf_tag_shape = (swf_tag_shape_detail_t *) tag->detail;
-        swf_tag_shape->_parse_condition = SWF_TAG_SHAPE_PARSE_CONDITION_BITMAP;
+//        swf_tag_shape->_parse_condition = SWF_TAG_SHAPE_PARSE_CONDITION_BITMAP;
         ret = swf_tag_shape_input_detail(tag, NULL);
         if (ret) {
             swf_tag_shape_destroy_detail(tag);
@@ -209,31 +288,72 @@ int swf_tag_shape_bitmap_replace_refcid(swf_tag_t *tag, int cid) {
     } else {
         swf_tag_shape = (swf_tag_shape_detail_t *) tag->detail;
     }
-    //
-    for (i = 0 ; i < swf_tag_shape->shape_with_style.styles.fill_styles.count ; i++) {
-        swf_fill_style_t *fill_style;
-        fill_style = &(swf_tag_shape->shape_with_style.styles.fill_styles.fill_style[i]);
-        if (fill_style == NULL) {
-            fprintf(stderr, "swf_tag_shape_bitmap_replace_refcid: fill_style == NULL i=%d\n", i);
-            return 1; // Illegal!!!
+    
+    if (tag->code == 46) { // DefineMorphShape
+        morph_shape_check = 1;
+        styles = &(swf_tag_shape->morph_shape_with_style.styles);
+        shape_records = &(swf_tag_shape->morph_shape_with_style.shape_records);
+//        shape_records = &(swf_tag_shape->morph_shape_with_style.shape_records_morph);
+    } else {
+        styles = &(swf_tag_shape->shape_with_style.styles);
+        shape_records = &(swf_tag_shape->shape_with_style.shape_records);
+    }
+    while (1) {
+        for (i = 0 ; i < styles->fill_styles.count ; i++) {
+            swf_fill_style_t *fill_style;
+            fill_style = &(styles->fill_styles.fill_style[i]);
+            if (fill_style == NULL) {
+                fprintf(stderr, "swf_tag_shape_bitmap_replace_refcid: fill_style == NULL i=%d\n", i);
+                return 1; // Illegal!!!
+            }
+            switch (fill_style->type) {
+              case 0x40: // tilled  bitmap fill with smoothed edges
+              case 0x41: // clipped bitmap fill with smoothed edges
+              case 0x42: // tilled  bitmap fill with hard edges
+              case 0x43: // clipped bitmap fill with hard edges
+                if (fill_style->bitmap.bitmap_ref == from_cid) {
+                    fill_style->bitmap.bitmap_ref = to_cid;
+                    if (tag->data) {
+                        // 内容が変わったので元データは削除
+                        free(tag->data);
+                        tag->data = NULL;
+                    }
+                    return 0; // success!
+                }
+                break;
+              default:
+                break;
+            }
+        }
+        // new style を探す
+        for ( ; shape_records ; shape_records = shape_records->next) {
+            if ((shape_records->first_6bits) && 
+                ((shape_records->first_6bits & 0x20) == 0)) {
+                if (shape_records->shape.shape_setup.shape_has_new_styles) {
+                    styles = &(shape_records->shape.shape_setup.styles);
+                    break;
+                }
+            }
         }
-        switch (fill_style->type) {
-          case 0x40: // tilled  bitmap fill with smoothed edges
-          case 0x41: // clipped bitmap fill with smoothed edges
-          case 0x42: // tilled  bitmap fill with hard edges
-          case 0x43: // clipped bitmap fill with hard edges
-            if (fill_style->bitmap.bitmap_ref != 0xffff) {
-                fill_style->bitmap.bitmap_ref = cid;
-                if (tag->data) {
-                    // 内容が変わったので元データは削除
-                    free(tag->data);
-                    tag->data = NULL;
+        shape_records = _swf_tag_shape_search_new_style_in_shape_records(shape_records);
+
+        if (shape_records) {
+            styles = &(shape_records->shape.shape_setup.styles);
+            shape_records = shape_records->next; // next
+        } else {
+            morph_shape_check = 0;
+            if (morph_shape_check) {        
+                shape_records = &(swf_tag_shape->morph_shape_with_style.shape_records);
+                shape_records = _swf_tag_shape_search_new_style_in_shape_records(shape_records);
+                if (shape_records) {
+                    styles = &(shape_records->shape.shape_setup.styles);
+                    morph_shape_check = 0;
+                } else {
+                    break; // finish
                 }
-                return 0; // success!
+            } else {
+                break; // finish
             }
-            break;
-          default:
-            break;
         }
     }
     return 1; // not found
@@ -247,6 +367,9 @@ swf_tag_shape_output_detail(swf_tag_t *tag, unsigned long *length,
     unsigned char *data;
     int ret;
     (void) swf;
+    long offset_of_offset_morph = 0;
+    long tmp_offset_byte = 0;
+    long tmp_offset_bit = 0;
     *length = 0;
     // build context
     swf_tag_shape->_current_fill_style_num = 0;
@@ -258,12 +381,12 @@ swf_tag_shape_output_detail(swf_tag_t *tag, unsigned long *length,
     bs = bitstream_open();
     bitstream_putbytesLE(bs, swf_tag_shape->shape_id, 2);
     swf_rect_build(bs, &(swf_tag_shape->rect));
-
+    
     // DefineMorphShape, DefineMorphShape2
-    swf_tag_shape->is_morph = (tag->tag == 46) || (tag->tag == 84);
+    swf_tag_shape->is_morph = (tag->code == 46) || (tag->code == 84);
     // DefineShape4, DefineMorphShape2
-    swf_tag_shape->has_strokes = (tag->tag == 83) || (tag->tag == 84);
-
+    swf_tag_shape->has_strokes = (tag->code == 83) || (tag->code == 84);
+    
     if (swf_tag_shape->is_morph) {
         ret = swf_rect_build(bs, &(swf_tag_shape->rect_morph));
         if (ret) {
@@ -281,19 +404,30 @@ swf_tag_shape_output_detail(swf_tag_t *tag, unsigned long *length,
         }
         if (swf_tag_shape->is_morph) {
             ret = swf_rect_build(bs, &(swf_tag_shape->stroke_rect_morph));
-	    if (ret) {
-	        fprintf(stderr, "ERROR: swf_tag_shape_input_detail: swf_tag_shape->stroke_rect_morph build failed\n");
-	        bitstream_close(bs);
-	        return NULL;
-	    }
+            if (ret) {
+                fprintf(stderr, "ERROR: swf_tag_shape_input_detail: swf_tag_shape->stroke_rect_morph build failed\n");
+                bitstream_close(bs);
+                return NULL;
+            }
         }
         bitstream_putbits(bs, 6, swf_tag_shape->define_shape_reserved );
         bitstream_putbits(bs, 1, swf_tag_shape->define_shape_non_scaling_strokes );
         bitstream_putbits(bs, 1, swf_tag_shape->define_shape_scaling_strokes);
     }
     if (swf_tag_shape->is_morph) {
-        bitstream_putbytesLE(bs, 4, swf_tag_shape->offset_morph);
+        bitstream_align(bs);
+        // 後で上書きするので、offset を覚えておく
+        offset_of_offset_morph = bitstream_getbytepos(bs);
+        bitstream_putbytesLE(bs, swf_tag_shape->offset_morph, 4);
         swf_morph_shape_with_style_build(bs, &swf_tag_shape->morph_shape_with_style, tag);
+        // offset_morph のフィールドに戻って上書きする
+        // offset_morph の後ろからの offset 差なので、- 4 する
+        swf_tag_shape->offset_morph = swf_tag_shape->morph_shape_with_style.offset_of_end_edges - offset_of_offset_morph - 4;
+        tmp_offset_byte = bitstream_getbytepos(bs); // offset を退避
+        tmp_offset_bit = bitstream_getbitpos(bs);
+        bitstream_setpos(bs, offset_of_offset_morph, 0);
+        bitstream_putbytesLE(bs, swf_tag_shape->offset_morph , 4);
+        bitstream_setpos(bs, tmp_offset_byte, tmp_offset_bit); // 元に戻す
     } else {
         ret = swf_shape_with_style_build(bs, &swf_tag_shape->shape_with_style, tag);
         if (ret) {
@@ -316,7 +450,7 @@ swf_tag_shape_print_detail(swf_tag_t *tag,
     print_indent(indent_depth);
     printf("shape_id=%d\n", swf_tag_shape->shape_id);
     swf_rect_print(&(swf_tag_shape->rect), indent_depth);
-    print_indent(indent_depth);
+//    print_indent(indent_depth);
 //    printf("is_morph=%d has_strokes=%d\n",
 //           swf_tag_shape->is_morph, swf_tag_shape->has_strokes);
     if (swf_tag_shape->is_morph) {
@@ -429,6 +563,8 @@ swf_tag_shape_apply_type_tilled(void *detail, int shape_id) {
         return 1;
     }
     if (shape_id != swf_tag_shape->shape_id) {
+        fprintf(stderr, "swf_tag_shape_apply_type_tilled: shape_id(%d) != shape->shape_id(%d)\n",
+	      shape_id, swf_tag_shape->shape_id);
         return 1;
     }
     count = swf_tag_shape->shape_with_style.styles.fill_styles.count;
